from dataclasses import dataclass, field
from collections import defaultdict
from typing import TYPE_CHECKING
from datetime import datetime

from .util import remote_addr

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(kw_only=True)
class ChannelStats:
    peeks: defaultdict[str, int] = field(default_factory=lambda: defaultdict(int))
    reads: defaultdict[str, int] = field(default_factory=lambda: defaultdict(int))
    writes: defaultdict[str, int] = field(default_factory=lambda: defaultdict(int))
    deletes: defaultdict[str, int] = field(default_factory=lambda: defaultdict(int))
    natime: datetime = field(default_factory=datetime.now)  # Last time a NEW read/write/delete occurred


@dataclass(kw_only=True)
class AdminStats:
    time: datetime = field(default_factory=datetime.now)
    version: str | None = None
    signer: Path | None = None
    uid_valid: bool = False
    uid: str | None = None
    command: str
    host: str


@dataclass(kw_only=True, frozen=True)  # Note that members are not frozen
class Stats:
    start: datetime = field(default_factory=datetime.now)
    channels: defaultdict[str, ChannelStats] = field(default_factory=lambda: defaultdict(ChannelStats))
    admin: list[AdminStats] = field(default_factory=list)

    def peek(self, channel: str) -> None:
        self._update(channel, "peeks")

    def read(self, channel: str) -> None:
        self._update(channel, "reads")

    def write(self, channel: str) -> None:
        self._update(channel, "writes")

    def delete(self, channel: str) -> None:
        self._update(channel, "deletes")

    def _update(self, channel: str, name: str) -> None:
        getattr(self.channels[channel], name)[remote_addr()] += 1
        self.channels[channel].natime = datetime.now()



from typing import TYPE_CHECKING, cast
from collections.abc import Callable
from collections import deque
from datetime import datetime
from logging import getLogger
from json import loads, dumps
from base64 import b85encode
from pathlib import Path
import json
import zlib

from cryptography.hazmat.primitives.serialization import load_ssh_private_key  # type: ignore[attr-defined]
from cryptography.exceptions import UnsupportedAlgorithm
from requests import Session

from ..shared import QueryResponse, AdminMessage, AdminEC, version
from .. import shared  # Let BLOCKED_EC be used in match statements
from .client import Config, UsageError, BlockedError

if TYPE_CHECKING:
    from requests import Response


ADMIN_REQUEST_TIMEOUT: int = 60
type _Signer = Callable[[bytes], bytes]
_LOG = "admin"


#
# Exceptions
#


class AccessDenied(RuntimeError):
    """
    Raised when the server denies an admin request
    """


#
# Main Classes
#


class _Methods:
    """
    Methods that can be called by the Admin class
    These methods may return sensitive data and should only be used with SSL
    Any public function may require SSL by invoke it with the prefix _VERIFY_SSL
    All public functions besides 'get' may access self._conf (it will not be None)
    """

    def __init__(self, sign: _Signer, conf: Config) -> None:
        self._log = getLogger(_LOG)
        self._uids: deque[str] = deque()
        self._session = Session()
        self._sign = sign
        self._conf = conf

    # Helpers

    def _require_ok(self, r: Response) -> None:
        if not r.ok:
            what = f"Error {r.status_code}: {r.text}"
            self._log.critical(what)
            raise RuntimeError(what)

    def _request(self, path: str, body: str = "") -> Response:
        """
        Send a request to the server
        """
        if len(self._uids) == 0:
            r = self._session.get(f"{self._conf.url}/admin/uid", timeout=ADMIN_REQUEST_TIMEOUT)
            self._require_ok(r)
            if r.status_code == shared.BLOCKED_EC:
                raise BlockedError()
            self._uids += r.json()
        uid: str = self._uids.popleft()
        # Sign and send POST request
        self._log.info("Signing request for path=%s with body=%s", path, body)
        msg = AdminMessage(path=path, body=body, uid=uid).bytes()
        data = b"\n".join((bytes(version), b85encode(self._sign(msg)), msg))
        ret = self._session.post(f"{self._conf.url}{path}", data=data, timeout=ADMIN_REQUEST_TIMEOUT)
        match ret.status_code:
            case shared.BLOCKED_EC:
                raise BlockedError()
            case AdminEC.unauthorized:
                self._log.critical("Admin access denied")
                raise AccessDenied()
            case AdminEC.illegal_version:
                raise UsageError(ret.text)
        self._require_ok(ret)
        assert not ret.status_code == AdminEC.invalid, "Sanity check failed"
        return ret

    # Non-SSL-Protected methods

    def debug(self) -> bool:
        """
        :return: True if the server is in debug mode, else False
        """
        return self._request("/admin/debug").text == "True"

    # SSL-Protected methods

    def log(self, output_file: Path | None = None) -> None:
        """
        Download the server log
        """
        out = zlib.decompress(self._request("/admin/log").content)
        if output_file is None:
            print(out.decode())
            return
        self._log.info("Writing log to %s", output_file)
        output_file.write_bytes(out)

    def log_level(self, level: int | str | None):
        old, new = self._request("/admin/log-level", "" if level is None else str(level)).text.split("\n")
        print(f"Log level: {old} -> {new}")

    def stats(self) -> None:
        """
        Give the client a bunch of stats about the server
        """
        print(dumps(loads(self._request("/admin/stats").text), indent=4))

    def channels(self) -> None:
        """
        Request information about the channels the server has
        """
        if not (raw := self._request("/admin/channels").json()):
            print("Server is empty")
            return
        for i in raw.values():
            i["expiration"] = datetime.fromisoformat(i["expiration"])
        data = {i: QueryResponse(**k) for i, k in raw.items()}
        mx = max(len(i) for i in data)
        print("\n".join(f"{i.ljust(mx)} : {k}" for i, k in data.items()))

    def _lock(self, lock_: bool) -> None:
        if not self._conf.channel:
            raise UsageError("Channel must be set to lock/unlock")
        print(self._request("/admin/lock", dumps({"channel": self._conf.channel, "lock": lock_})).text)

    def lock(self) -> None:
        """
        Lock a channel
        """
        self._lock(True)

    def unlock(self) -> None:
        """
        Unlock a channel
        """
        self._lock(False)

    def _block(self, name: str, block: list[str] | None, unblock: list[str] | None) -> None:
        if block is None and unblock is None:
            blocked = self._request(f"/admin/{name}", f'{{"{name}": null}}').text
            blocked = json.dumps(json.loads(blocked), indent=4)
            print(f"Blocked {name}s: {blocked}")
            return
        for lst, ban in ((block, True), (unblock, False)):
            if lst is None:
                continue
            for obj in lst:
                self._request(f"/admin/{name}", dumps({name: obj, "block": ban}))
                print(f"{"" if ban else "UN"}BLOCKED: {obj}")

    def ip(self, block: list[str] | None, unblock: list[str] | None) -> None:
        """
        Request the blocked ip addresses, or block / unblock an ip address
        """
        self._block("ip", block, unblock)

    def route(self, block: list[str] | None, unblock: list[str] | None) -> None:
        """
        Request the blocked routes, or block / unblock a route
        """
        fix = lambda b: None if b is None else [i if i.startswith("/") else f"/{i}" for i in b]
        self._block("route", fix(block), fix(unblock))


class Admin:
    """
    A class used to ask the server to run admin functions
    """

    def __init__(self, conf: Config):
        self._log = getLogger(_LOG)
        self._ssl: bool = any(i in conf.url for i in ("https", ":443/"))
        if not conf.url or not conf.key_file:
            raise UsageError("Admin mode requires a URL and key-file to be set")
        self._methods = _Methods(self._load_ssh_key_file(conf.key_file), conf)

    def _load_ssh_key_file(self, key_file: Path) -> _Signer:
        """
        Load a private key from a file
        :return: A function that can sign data using the key file
        """
        self._log.info("Extracting private key from %s", key_file)
        if not key_file.exists():
            raise UsageError(f"Key file {key_file} does not exist")
        try:
            key = load_ssh_private_key(key_file.read_bytes(), None)
        except UnsupportedAlgorithm as e:
            raise UsageError(f"Key file {key_file} is not a supported ssh key") from e
        if not hasattr(key, "sign"):
            raise UsageError(f"Key file {key_file} does not support signing")
        return cast(_Signer, key.sign)

    def __getitem__(self, item: str) -> Callable[..., None]:
        """
        Get the desired admin function
        """
        if item.startswith("_"):
            raise KeyError(f"Admin method {item} is private")
        debug = self._methods.debug()
        if item == "debug":
            return lambda: print(f"Server is running in {'DEBUG' if debug else 'RELEASE'} mode")
        if not debug and not self._ssl:
            raise RuntimeError("Refusing to send admin request to server in release mode over plaintext")
        return getattr(self._methods, item)



from typing import TYPE_CHECKING, NamedTuple
from logging import getLogger
import hashlib

from Cryptodome.Random import get_random_bytes
from Cryptodome.Cipher import AES

from ...shared import TRACE, LFS

if TYPE_CHECKING:
    from collections.abc import Callable


class _EncryptedData(NamedTuple):
    text: bytes
    salt: bytes
    nonce: bytes
    tag: bytes

    def encode(self) -> bytes:
        line1 = b" ".join(str(len(i)).encode() for i in self) + b"\n"  # pylint: disable=not-an-iterable
        return line1 + b"".join(self)

    @classmethod
    def decode(cls, raw: bytes) -> list[_EncryptedData]:  # typing.Self in python3.11
        ret: list[_EncryptedData] = []  # python3.11 typing.Self
        end: int = 0
        while end < len(raw):  # We use this loop to avoid constantly slicing big data
            parts: list[bytes] = []
            start = raw.index(b"\n", end) + 1
            for i in (int(k.decode()) for k in raw[end : start - 1].split(b" ")):
                end = start + i
                parts.append(raw[start:end])
                start += i
            if len(parts) != len(cls._fields):
                raise ValueError("Bad encrypted data")
            ret.append(cls(*parts))
        return ret


def _aes(salt: bytes, password: str, nonce: bytes | None = None):
    scrypt = hashlib.scrypt(salt=salt, password=password.encode(), n=2**14, r=8, p=1, dklen=32)
    return AES.new(scrypt, AES.MODE_GCM, nonce=nonce)


def encrypt(data: bytes, compress: Callable[[bytes], bytes], password: str | None) -> bytes:
    log = getLogger("encrypt")
    if not password or not data:
        log.log(TRACE, "Not encrypting")
        return data
    log.debug("Compressing %s chunk", LFS(data))
    data = compress(data)
    log.debug("Encrypting compressed %s chunk", LFS(data))
    salt = get_random_bytes(AES.block_size)
    aes = _aes(salt, password)
    text, tag = aes.encrypt_and_digest(data)
    return _EncryptedData(text, salt, aes.nonce, tag).encode()


def decrypt(data: bytes, decompress: Callable[[bytes], bytes], password: str | None) -> bytes:
    log = getLogger("decrypt")
    if not password or not data:
        log.log(TRACE, "Not decrypting")
        return data
    log.debug("Extracting chunks from %s of data", LFS(data))
    es = _EncryptedData.decode(data)
    sfx = "s" if len(es) != 1 else ""
    log.debug("Decrypting %d chunk%s", len(es), sfx)
    r = [_aes(e.salt, password, e.nonce).decrypt_and_verify(e.text, e.tag) for e in es]
    log.debug("Decompressing %d chunk%s", len(es), sfx)
    r = [decompress(i) for i in r]
    if len(es) > 1:
        log.debug("Merging chunks")
    return r[0] if len(r) == 1 else b"".join(r)



from functools import total_ordering

from .. import __version__


@total_ordering
class Version:
    """
    A class to represent a version number
    Implements comparison, equality, and str operators
    Allows invalid versions but marks them as such
    """

    _invalid = (-1, -1, -1)
    _invalid_str = "Unable to parse version"

    def __init__(self, v: str | bytes):
        try:
            self.str: str = v if isinstance(v, str) else v.decode()
            tup = tuple(int(i) for i in self.str.split("."))
            self.tuple = tup if len(tup) == 3 else self._invalid
        except ValueError:
            self.str = self._invalid_str
            self.tuple = self._invalid

    def invalid(self) -> bool:
        return self.tuple == self._invalid

    def __str__(self) -> str:
        return self.str

    def __bytes__(self):
        return str(self).encode()

    def __lt__(self, other: Version):
        return self.tuple < other.tuple

    def __eq__(self, other: object):
        return isinstance(other, Version) and self.str == other.str


version = Version(__version__)

WEB_VERSION = Version("0.0.0")
assert not WEB_VERSION.invalid()  # nosec B101



from typing import TYPE_CHECKING, Self
from logging import getLogger

from tqdm.contrib.logging import logging_redirect_tqdm
import tqdm

from .delete import delete
from .data import Result

if TYPE_CHECKING:
    from typing import IO
    from .data import Config, Mode


class Progress:
    """
    A small class to handle progress bars and progression data of send/recv
    """

    DOF_EXC = KeyboardInterrupt | Exception

    __slots__ = ("result", "dof", "_config", "_redir", "_pbar")

    def __init__(self, config: Config, mode: Mode):
        self.result = Result(total=mode.total, checksum=mode.checksum)
        self.dof: bool = False
        self._config = config
        self._redir = logging_redirect_tqdm()
        self._pbar = tqdm.tqdm(
            desc="Uploading" if mode.send else "Downloading",
            disable=not bool(mode.progress),
            # pylint: disable=unidiomatic-typecheck
            total=mode.progress if type(mode.progress) is int else None,
            dynamic_ncols=True,
            leave=False,
            unit_divisor=1000,
            unit_scale=True,
            unit="B",
        )

    def __enter__(self) -> Self:
        self._redir.__enter__()
        self._pbar.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.dof and isinstance(exc_val, self.DOF_EXC):
            log = getLogger("Progress")
            log.warning("Caught %s. Deleting channel: %s", type(exc_val), self._config.channel)
            delete(self._config)
        r1 = bool(self._pbar.__exit__(exc_type, exc_val, exc_tb))
        r2 = bool(self._redir.__exit__(exc_type, exc_val, exc_tb))
        return r1 or r2

    def update(self, data: bytes, *, file: IO[bytes] | None = None) -> None:
        self.dof = True
        self._pbar.update(len(data))
        if self.result.total is not None:
            self.result.total += len(data)
        if self.result.checksum is not None:
            self.result.checksum.update(data)
        if file is not None:
            file.write(data)
            file.flush()



from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING
from logging import getLogger
from pathlib import Path
from time import sleep
from sys import stdout
import tarfile
import atexit

from zstandard import ZstdDecompressor

from ...shared import (
    DownloadRequestParams,
    DownloadResponseHeaders,
    DownloadEC,
    SpooledTempFile,
    version,
    LFS,
)
from .errors import MultipleClients, ChannelLocked, ReportThis, VersionError, StreamError, NoData
from .util import wait_delay_sec, request
from .progress import Progress
from .crypt import decrypt

if TYPE_CHECKING:
    from typing import IO
    from requests import Response
    from .data import Result, Config, Mode


_LOG = "recv"


# pylint: disable=too-many-arguments
def _recv_error_helper(r: Response, conf: Config, peek: bool, put: bool, waited: bool) -> None:
    """
    Raise an exception according to the recv response error
    """
    match r.status_code:
        case DownloadEC.wrong_version:
            v = r.text.split(":")[-1].strip()
            raise VersionError(f"Version mismatch; uploader version = {v}; force a read with --force")
        case DownloadEC.illegal_version:
            raise VersionError(r.text)
        case DownloadEC.no_data:
            if put:
                msg = "This data stream no longer exists; maybe the sender cancelled sending?"
                raise MultipleClients(msg)
            raise NoData(f"The channel {conf.channel} is empty.")
        case DownloadEC.conflict:
            if put:
                raise MultipleClients("This data stream no longer exists; maybe the channel was deleted?")
            raise ReportThis(r.text)
        case DownloadEC.cannot_peek:
            msg = "Too much data to peek; data is being streamed and does not all exist on server."
            raise StreamError(msg)
        case DownloadEC.in_use:
            if peek and waited:
                raise MultipleClients("Another client started reading the data before peek was complete")
            raise MultipleClients(r.text)
        case DownloadEC.locked:
            raise ChannelLocked(r.text)
        case DownloadEC.forbidden:
            raise ReportThis("Attempt to read from stream with stream ID.")
        case _:
            raise RuntimeError(f"Error {r.status_code}", r.text)


def _recv_error(*args, **kwargs) -> None:
    """
    Wrapper for _recv_error_helper
    """
    try:
        _recv_error_helper(*args, **kwargs)
    except (VersionError, NoData, ReportThis, MultipleClients, RuntimeError) as e:
        getLogger(_LOG).error(", ".join(e.args))
        raise e


# pylint: disable=too-many-positional-arguments
def _recv_body(
    conf: Config,
    progress: Progress,
    block: bool,
    peek: bool,
    params: DownloadRequestParams,
    lvl: int,
    file: IO[bytes],
) -> int | None:
    log = getLogger(_LOG)
    decompress = ZstdDecompressor().decompress
    r = request("GET", conf.channel_url(), params=params.to_dict(), timeout=conf.timeout)
    if r.ok:
        progress.dof = True
        headers = DownloadResponseHeaders.from_dict(r.headers)
        log.info("Received %s", LFS(r.content))
        got: bytes = decrypt(r.content, decompress, conf.password if headers.encrypted else None)
        progress.update(got, file=file)
        if headers.final:
            return None  # Stream complete
        params.stream_id = headers.stream_id
        return 0
    if (block and r.status_code == DownloadEC.no_data) or (r.status_code == DownloadEC.wait):
        delay = wait_delay_sec(lvl)
        log.info("No data available yet, sleeping for %s second(s)", delay)
        sleep(delay)
        return lvl + 1
    log.error("Error reading from channel %s. Status Code: %s", conf.channel, r.status_code)
    _recv_error(r, conf, peek, params.stream_id is not None, lvl != 0)
    raise NotImplementedError("Unreachable code")


def _recv(conf: Config, mode: Mode, file: IO[bytes]) -> Result:
    """
    Receive data from the remote pipe and output to file
    """
    block: bool = mode.block
    log = getLogger(_LOG)
    log.info("Reading from channel %s with peek=%s and force=%s", conf.channel, mode.peek, mode.force)
    params = DownloadRequestParams(version=version, override=mode.force, delete=not mode.peek)
    lvl: int | None = 0
    try:
        with Progress(conf, mode) as progress:
            while lvl is not None:
                if (lvl := _recv_body(conf, progress, block, mode.peek, params, lvl, file)) == 0:
                    block = False  # Stop blocking after first successful read
    except BrokenPipeError:
        log.warning("BrokenPipeError: output pipe closed")
    else:
        log.info("Stream complete")
    return progress.result


def _rm_existing(path: Path, yes: bool) -> None:
    if not path.exists(follow_symlinks=False):
        return
    if not yes:
        raise FileExistsError(f"Path {path} already exists; to overwrite use --yes")
    (path.rmdir if path.is_dir(follow_symlinks=False) else path.unlink)()


def recv(conf: Config, mode: Mode) -> Result:
    """
    Receive data from the remote pipe
    """
    log = getLogger(_LOG)

    # Check for existing paths; if mode.dir, set mode.file (untar it later)
    out_f: IO[bytes] = stdout.buffer
    if mode.file is not None:
        _rm_existing(mode.file, mode.yes)
        out_f = mode.file.open("xb")
    if mode.dir is not None:
        _rm_existing(mode.dir, mode.yes)
        log.debug("Creating placeholder directory %s", mode.file)
        mode.dir.mkdir(mode=0)  # Optional; it is just so that no one create anything here while we work
        atexit.register(mode.dir.rmdir)
        out_f = SpooledTempFile()

    # Recv data into mode.file
    try:
        ret = _recv(conf, mode, out_f)
    # Cleanup
    except Exception:
        if mode.file is not None:
            empty: bool = out_f.tell() == 0
            out_f.close()
            if empty:  # Only cleanup output file if it is empty
                mode.file.unlink()
        raise
    # Return if not dir
    if mode.dir is None:
        if mode.file is not None:
            out_f.close()
        return ret

    # Unpack tarball
    out_f.seek(0)
    with TemporaryDirectory(suffix=f" {mode.dir.name}") as d:
        temp_d = Path(d)
        log.debug("Extracting to: %s", temp_d)
        with tarfile.open(fileobj=out_f) as tb:
            log.info("Extracting with filter=tar_filter for safety")
            tb.extractall(path=temp_d, filter=tarfile.tar_filter)  # nosec
        out_f.close()
        # Since our tarball.add -> tarball.extract adds an enclosing dir, pull the child out
        log.debug("Removing placeholder dir %s", mode.dir)
        mode.dir.rmdir()  # Remove placeholder
        atexit.unregister(mode.dir.rmdir)  # Do not delete anymore
        child: Path | None = None
        if len(children := list(temp_d.glob("*"))) == 1 and children[0].is_dir(follow_symlinks=False):
            log.info("Only one directory found in tarball, will rename as enclosing directory")
            child = children[0]
        log.info("Moving unpacked dir into place: %s", mode.dir)
        (child if child is not None else temp_d).rename(mode.dir)
    return ret



from logging import getLogger


class UsageError(ValueError):
    """
    Raised when the user used the client incorrectly (ex. CLI args)
    """

    def __init__(self, msg: str):
        # This is likely CRITICAL, but also likely to be printed out anyway, so log it as info
        getLogger("UsageError").info("%s", msg)
        super().__init__(msg)


class BlockedError(UsageError):
    """
    Raised when the server has blocked this IP address
    """

    def __init__(self) -> None:
        super().__init__("This IP address is blocked by the server")


class VersionError(UsageError):
    """
    Raised when the server has rejected the client version
    """


class ChannelLocked(RuntimeError):
    """
    Raised when attempting to modify a channel is locked
    """


class NoData(ValueError):
    """
    Raised when there is no data available on a channel
    """


class StreamError(RuntimeError):
    """
    Raised when an action fails due to the data being
    streamed in chunks rather than uploaded all at once
    """


class MultipleClients(StreamError):
    """
    Raised when failure occurs due to another user using the same channel
    """


class ReportThis(RuntimeError):
    """
    A RuntimeError that should be reported
    """

    def __init__(self, msg: str):
        super().__init__(f"{msg}\nPlease report this.")



from logging import getLogger, INFO
from typing import TYPE_CHECKING
from dataclasses import asdict
from collections import deque
from threading import RLock
import json

from ...shared import Stats, Version, restrict_umask, version
from .stream import Stream

if TYPE_CHECKING:
    from typing import BinaryIO
    from pathlib import Path


MIN_SAVE_STATE_VERSION = Version("8.1.0")


def _writeline(f: BinaryIO, s: bytes):
    f.write(str(len(s)).encode() + b"\n")
    f.write(s)
    f.write(b"\n")


def _readline(f: BinaryIO) -> bytes:
    size = int(f.readline().strip())
    ret = b""
    while len(ret) < size:
        ret += f.read(size - len(ret))
    if f.read(1) != b"\n":
        raise ValueError("Expected newline")
    return ret


class ServerShutdown(RuntimeError):
    """
    Raised when trying to acquire the lock on a server that is already shut down
    """


class UnlockedState:
    """
    A class that holds the state of a server
    This class is not thread safe and access to it should be protected by a lock
    """

    __slots__ = ("streams", "shutdown", "stats")

    _log = getLogger("UnlockedState")

    def __init__(self) -> None:
        self.streams: dict[str, Stream] = {}
        self.shutdown: bool = False
        self.stats = Stats()

    def load(self, file: Path) -> None:
        """
        Save the state of the server (does not load stats)
        """
        if len(self.streams):
            self._log.error("Existing state detected; will not overwrite")
            raise RuntimeError("Do not load a state on top of an existing state")
        if not file.exists():
            self._log.warning("State file %s not found. State is set to empty", file)
            return
        if not self._load(file):
            self._log.warning("Failed to load saved state. State is set to empty.")
            return
        self._log.debug("Creating server Stats")
        self.stats = Stats()
        _ = [self.stats.channels[i] for i in self.streams]
        self._log.info("State loaded successfully")

    def save(self, file: Path) -> None:
        """
        Save the program state (does not save stats)
        Do not call this unless the server is shutdown!
        Assumes self.RLock is acquired
        """
        if not self.shutdown:
            raise RuntimeError("Do save state before shutdown")
        if file.exists():
            self._log.info("Purging old program state...")
            file.unlink()
        self._save(file)
        if self._log.isEnabledFor(INFO):
            self._log.info("Channels saved: %s", ", ".join(self.streams.keys()))
        self._log.info("State saved successfully")

    def _save(self, file: Path) -> None:
        self._log.info("Saving state to: %s", file)
        with restrict_umask(0o6):
            with file.open("wb") as f:
                f.write(bytes(version) + b"\n")
                _writeline(f, str(len(self.streams)).encode())
                for name, s in self.streams.items():
                    d = asdict(s)
                    deq = d.pop("data")
                    _writeline(f, f"{name} {len(deq)} ".encode() + json.dumps(d, default=str).encode())
                    for i in deq:
                        _writeline(f, i)

    def _load(self, file: Path) -> bool:
        self._log.info("Loading %s", file)
        self.streams = {}
        with file.open("rb") as f:
            if (ver := Version(f.readline()[:-1])) < MIN_SAVE_STATE_VERSION:
                self._log.error("State version too old: %s", ver)
                return False
            for _ in range(int(_readline(f))):
                main = _readline(f).split(b" ", 2)
                body = json.loads(main[2])
                body["version"] = Version(body["version"])
                body["data"] = deque(_readline(f) for _2 in range(int(main[1])))
                self.streams[main[0].decode()] = Stream(**body)
        return True


class State:
    """
    A thread safe wrapper for ServerState
    """

    __slots__ = ("_lock", "_log", "_state", "_debug")

    def __init__(self, debug: bool) -> None:
        self._lock = RLock()
        self._log = getLogger("State")
        self._state = UnlockedState()
        self._debug: bool = debug

    @property
    def debug(self) -> bool:
        return self._debug

    def __enter__(self) -> UnlockedState:
        """
        Acquire the lock and return the state; will fail if the server is shutdown
        """
        self._lock.acquire()
        if self._state.shutdown:
            self._log.error("Lock acquired, but server is shut down")
            self._lock.release()
            raise ServerShutdown()
        return self._state

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()



from typing import TYPE_CHECKING
from logging import getLogger

from ...shared import DeleteEC
from .errors import ChannelLocked
from .util import request

if TYPE_CHECKING:
    from .data import Config


def delete(conf: Config) -> None:
    """
    Delete the channel
    """
    getLogger("delete").info("Deleting channel %s", conf.channel)
    r = request("DELETE", conf.channel_url(), timeout=conf.timeout)
    if r.status_code == DeleteEC.locked:
        raise ChannelLocked(r.text)
    if not r.ok:
        raise RuntimeError(r)



from logging import WARNING, INFO, DEBUG
from typing import TYPE_CHECKING
import logging

from human_readable import file_size

from .util import total_len

if TYPE_CHECKING:
    from collections.abc import Sequence


CF_KWARGS = {
    "fmt": "%(cute_asctime)s.%(msecs)03d - %(cute_levelname)s - %(cute_name)s - %(cute_message)s",
    "datefmt": "%H:%M:%S",
    "cute_widths": {
        "cute_levelname": 8,
        "cute_name": 13,
    },
}

TRACE = DEBUG - 5
assert TRACE > 0
_VERBOSITY: dict[int, int] = {0: WARNING, 1: INFO, 2: DEBUG, 3: TRACE}


class LFS:
    """
    Human-readable number of bytes, lazily evaluated for logging
    """

    def __init__(self, x: int | bytes | Sequence[bytes]) -> None:
        self._x = x if isinstance(x, int) else (total_len(x) if isinstance(x, list) else len(x))

    def __str__(self) -> str:
        return file_size(self._x)


def level(verbosity: int) -> int:
    return _VERBOSITY[max(i for i in _VERBOSITY if i <= verbosity)]


def define_trace():
    """
    Add a TRACE level to the logging module
    """
    assert not hasattr(logging, "TRACE"), "Already added TRACE level"
    logging.addLevelName(TRACE, "TRACE")



from typing import TYPE_CHECKING
from logging import getLogger
from functools import cache

from requests import Session, Request

from .errors import BlockedError
from ...shared import BLOCKED_EC

if TYPE_CHECKING:
    from requests import Response


_DEFAULT_TIMEOUT: int = 60
_WAIT_DELAY_SEC: dict[int, float] = {0: 0.3, 1: 0.5, 5: 1.0, 60: 2.0, 300: 5.0}


def wait_delay_sec(lvl: int) -> float:
    """
    :return: The number of seconds to wait before retrying a request
    """
    if lvl < 0:
        raise ValueError("Invalid level")
    return _WAIT_DELAY_SEC[max(i for i in _WAIT_DELAY_SEC if i <= lvl)]


@cache
def _session() -> Session:
    return Session()


def request(*args, timeout: int | None, **kwargs) -> Response:
    r = Request(*args, **kwargs).prepare()
    if r.body:
        getLogger("request").debug("Making %s request with %d bytes of data", r.method, len(r.body))
    ret = _session().send(r, timeout=_DEFAULT_TIMEOUT if timeout is None else timeout)
    if ret.status_code == BLOCKED_EC:
        raise BlockedError()
    return ret



from logging import getLevelNamesMapping, getLevelName, getLogger
from typing import TYPE_CHECKING, Any
from dataclasses import asdict
from json import loads
import zlib

from flask import Response

from ...shared import AdminEC
from ..util import plaintext, json_response
from .verify import Verify

if TYPE_CHECKING:
    from pathlib import Path
    from ..app import Blocked
    from ..server import State


_UIDS_PER_QUERY: int = 2


class Methods:
    """
    Protected methods that can be accessed by the Admin class
    """

    def __init__(self, log_file: Path | None, blocked: Blocked) -> None:
        self._log = getLogger("Admin")
        self._log_file = log_file
        self._blocked = blocked
        self._log.debug("Log file set to %s", log_file)

    @staticmethod
    def debug(state: State, _: str) -> Response:
        return plaintext(str(state.debug))

    def log(self, *_) -> Response:
        for i in getLogger().handlers:
            i.flush()
        if self._log_file is None:
            return Response("Missing log file", status=500, mimetype="text/plain")
        data = zlib.compress(self._log_file.read_bytes().strip())
        self._log.debug("Sending compressed log of size: %s", len(data))
        return Response(data, status=200, mimetype="application/octet-stream")

    def log_level(self, state: State, body: str) -> Response:
        root = getLogger()
        new = (old := getLevelName(root.getEffectiveLevel()))
        if body:
            try:
                new = getLevelName(lvl := int(getLevelNamesMapping().get(body.upper(), body)))
                self._log.info("Setting log level to %s", new)
                root.setLevel(lvl)
                if state.debug:
                    getLogger("werkzeug").setLevel(lvl)
                self._log.debug("Log level set to %s", new)
            except ValueError:
                return Response(f"Invalid log level: {body}", status=AdminEC.invalid, mimetype="text/plain")
        return Response(f"{old}\n{new}", status=200, mimetype="text/plain")

    @staticmethod
    def stats(state: State, _: str, blocked: dict[str, list[list[str]]]) -> Response:
        with state as s:
            stats = asdict(s.stats)
        return json_response({"server": stats, "blocked": blocked})

    @staticmethod
    def channels(state: State, _: str) -> Response:
        """
        Return a list of the server's current channels and stats
        """
        with state as s:
            output = {i: asdict(k.query()) for i, k in s.streams.items()}
        return json_response(output)

    def lock(self, state: State, body: str) -> Response:
        js = loads(body.strip())
        with state as unlocked:
            if (s := unlocked.streams.get(channel := js["channel"], None)) is None:
                return Response(f"Channel {channel} not found", status=AdminEC.invalid)
            lock_s = f"{('' if (lock := js['lock']) else 'UN')}LOCKED"
            self._log.info("Setting channel %s to %s", channel, lock_s)
            s.locked = lock
        return Response(f"Channel {channel} is now {lock_s}", status=200)

    def _block(self, name: str, body: str) -> Response:
        js = loads(body.strip())
        with self._blocked as data:
            if (obj := js[name]) is None:
                return json_response(list(getattr(data, f"{name}_blacklist")))
            lst = getattr(data, f"{name}_blacklist")
            if js["block"]:
                if obj not in lst:
                    self._log.info("Blocking %s: %s", name, obj)
                    (lst.add if isinstance(lst, set) else lst.append)(obj)
            elif obj in lst:
                while obj in lst:
                    self._log.info("Unblocking %s: %s", name, obj)
                    lst.remove(obj)
        return Response(status=200)

    def ip(self, _: State, body: str) -> Response:
        return self._block("ip", body)

    def route(self, _: State, body: str) -> Response:
        return self._block("route", body)


class Admin:
    """
    Admin class that protects the server from unauthorized access
    It requires a valid signature to access any method from the Methods class
    Any public method from the method class is a member of this class, wrapped in a signature verification
    Signed requests must contain a valid signed UID that is only valid for a short period of time
    """

    __slots__ = ("_verify", "_methods")

    def __init__(self, log_file: Path, key_files: list[Path], blocked: Blocked) -> None:
        self._verify = Verify(key_files)
        self._methods = Methods(log_file, blocked)

    def __getattr__(self, item: str) -> Any:
        """
        Override the getattribute method to wrap most public members with signature verification
        """
        if item.startswith("_"):
            raise AttributeError(f"{item} is a private member")

        def wrapper(state: State, *args, **kwargs) -> Response:
            assert self._verify is not None, "Admin not initialized"
            if isinstance(rv := self._verify(item, state), str):
                return getattr(self._methods, item)(state, rv, *args, **kwargs)
            return rv

        return wrapper

    def uids(self) -> Response:
        """
        Get a few UIDSs that may each be used in a signature to access the server exactly once
        """
        return json_response(self._verify.uid.new(_UIDS_PER_QUERY))



from typing import TYPE_CHECKING, cast
from collections import deque
from logging import getLogger

from flask import Response, request

from ...shared import (
    WEB_VERSION,
    DownloadResponseHeaders,
    DownloadRequestParams,
    DownloadEC,
    total_len,
    TRACE,
    LFS,
)
from ..util import MIN_VERSION, MAX_SIZE_SOFT, plaintext
from .util import log_response, log_params

if TYPE_CHECKING:
    from collections.abc import Sequence
    from ..server import Stream
    from ..server import State


_LOG: str = "read"


def _check_if_aio(s: Stream, args: DownloadRequestParams) -> Response | None:
    if not args.delete or args.version == WEB_VERSION:
        mode = "web client" if args.delete else "peek"
        if args.stream_id is not None:
            return plaintext("Stream ID not allowed when using {mode}.", DownloadEC.forbidden)
        if not s.new:
            return plaintext("Another client has already connected to this pipe.", DownloadEC.in_use)
        if not s.upload_complete:
            if s.full():
                msg = f"Must wait until uploader completes upload when using {mode}"
                return plaintext(msg, DownloadEC.wait)
            msg = f"Too much data to read all at once: when using {mode}; data can only be read all at once."
            return plaintext(msg, DownloadEC.cannot_peek)
    return None


# pylint: disable=too-many-return-statements
def _read_error_check(s: Stream | None, args: DownloadRequestParams) -> Response | None:
    """
    :return: A response if the data in s should not be returned due to an error, else None
    """
    # No data found?
    if s is None:
        return plaintext("This channel is currently empty", DownloadEC.no_data)
    # If data must be all at once, handle it
    if err := _check_if_aio(s, args):
        return err
    # Stream ID check
    if args.stream_id is None and s.new is False:
        return plaintext("Another client has already connected to this pipe.", DownloadEC.in_use)
    if args.stream_id is not None and args.stream_id != s.id_:
        return plaintext("Stream ID mismatch", DownloadEC.conflict)
    # Web version cannot handle encryption
    if args.version == WEB_VERSION and s.encrypted:
        return plaintext("Web version cannot read encrypted data. Use the CLI: pip install rpipe", 422)
    # Version comparison; bypass if web version or override requested
    if args.version not in (WEB_VERSION, s.version) and not args.override:
        return plaintext(f"Override = False. Version should be: {s.version}", DownloadEC.wrong_version)
    # Not data currently available
    if not s.upload_complete and not s.data:
        return plaintext("No data available; wait for the uploader to send more", DownloadEC.wait)
    # Lock check
    if s.locked and args.delete:
        return plaintext("This channel is locked and cannot be edited; consider --peek", DownloadEC.locked)
    return None


@log_response(_LOG)
def read(state: State, channel: str) -> Response:
    """
    Get the data from channel, delete it afterward if required
    If web version: Fail if not encrypted, bypass version checks
    Otherwise: Version check
    """
    log = getLogger(_LOG)
    args = DownloadRequestParams.from_dict(request.args)
    log_params(log, args)
    if args.version != WEB_VERSION and (args.version < MIN_VERSION or args.version.invalid()):
        return plaintext(f"Bad version. Server requires >= {MIN_VERSION}", DownloadEC.illegal_version)
    with state as u:
        s: Stream | None = u.streams.get(channel, None)
        if (err := _read_error_check(s, args)) is not None:
            return err
        if TYPE_CHECKING:
            s = cast(Stream, s)  # For type checker
        # Read all at once if required
        if not args.delete:  # Peek mode (could also be web version)
            log.debug("Reading channel %s in peek mode", channel)
            u.stats.peek(channel)
            rdata: Sequence[bytes] = s.data
            final = True
        elif args.version == WEB_VERSION:
            log.debug("Reading channel %s from WEB_VERSION", channel)
            u.stats.read(channel)
            rdata = s.data
            s.data = deque()
            final = True
        # Standard read mode
        else:
            log.log(TRACE, "Reading channel %s in standard mode", channel)
            if s.new:
                s.new = False
                u.stats.read(channel)
            rdata = [s.data.popleft()] if s.data else []  # Ensure at least one packet if available
            while s.data and (len(s.data[0]) + total_len(rdata)) < MAX_SIZE_SOFT:
                rdata.append(s.data.popleft())
            final = s.upload_complete and not s.data
        if args.delete and final:
            log.debug("Channel %s empty and final; removing", channel)
            del u.streams[channel]
    log.log(TRACE, "Merging %d piece(s) of data; total length: %s", len(rdata), LFS(rdata))
    headers = DownloadResponseHeaders(encrypted=s.encrypted, stream_id=s.id_, final=final).to_dict()
    return Response(b"".join(rdata), mimetype="application/octet-stream", headers=headers)



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import TYPE_CHECKING
import random
import string

from ...shared import QueryResponse, total_len

if TYPE_CHECKING:
    from ...version import Version
    from collections import deque


CHARSET = string.ascii_lowercase + string.ascii_uppercase + string.digits
_PIPE_MAX_BYTES: int = 10**9


def _uid() -> str:
    return "".join(random.choices(CHARSET, k=32))  # nosec B311


@dataclass(kw_only=True)
class Stream:  # pylint: disable=too-many-instance-attributes
    """
    Holds data about a stream
    """

    ttl: int
    data: deque[bytes]
    upload_complete: bool  # If more data will be added
    new: bool = True  # If no data has been read
    locked: bool = False  # May only be set by admin
    # Constants
    encrypted: bool
    version: Version
    capacity: int = _PIPE_MAX_BYTES
    id_: str = field(default_factory=_uid)

    def __post_init__(self) -> None:
        self.expire: datetime  # Set by __setattr__
        self._capacity: int = _PIPE_MAX_BYTES
        self._constants = ("encrypted", "version", "id_", "_constants")

    def __setattr__(self, key, value):
        """
        __setattr__ override to prevent changing constants and updates self.expire
        """
        if key == "expire":
            raise AttributeError("Expiration date is automatically set; do not change it manually")
        if key in getattr(self, "_constants", {}):
            raise AttributeError("Cannot change constant values")
        super().__setattr__(key, value)
        if hasattr(self, "ttl"):  # hasattr b/c we might not during init
            # pylint: disable=attribute-defined-outside-init
            super().__setattr__("expire", datetime.now() + timedelta(seconds=self.ttl))

    def expired(self) -> bool:
        """
        Return true if the stream is expired and unlocked
        """
        return not self.locked and self.expire < datetime.now()

    def __len__(self) -> int:
        """
        :return: The number of bytes in the pipe
        """
        return total_len(self.data)

    def full(self) -> bool:
        """
        :return: True if the server pipe is full, else False
        """
        return len(self) >= self._capacity

    def query(self) -> QueryResponse:
        return QueryResponse(
            new=self.new,
            upload_complete=self.upload_complete,
            packets=len(self.data),
            size=len(self),
            encrypted=self.encrypted,
            version=self.version,
            expiration=self.expire,
            locked=self.locked,
        )



from datetime import datetime, timedelta
from logging import getLogger
from threading import RLock
from os import urandom


class UID:
    """
    A class to manage UIDs that are used for signature verification
    """

    _UID_EXPIRE: int = 300
    _UID_LEN: int = 32

    def __init__(self) -> None:
        self._uids: dict[str, datetime] = {}
        self._log = getLogger("UID")
        self._lock = RLock()

    def new(self, n: int) -> list[str]:
        ret = [urandom(self._UID_LEN).hex() for i in range(n)]
        with self._lock:
            eol = datetime.now() + timedelta(seconds=self._UID_EXPIRE)
            self._uids.update({i: eol for i in ret})
        self._log.debug("Generated %s new UIDs", n)
        return ret

    def verify(self, uid: str) -> bool:
        self._log.debug("Verifying UID: %s", uid)
        with self._lock:
            if uid not in self._uids:
                self._log.error("UID not found: %s", uid)
                return False
            if datetime.now() > self._uids.pop(uid):
                self._log.warning("UID expired: %s", uid)
                return False
            self._log.debug("UID verified")
        return True



from dataclasses import dataclass, asdict
from typing import TYPE_CHECKING, TypeVar
from datetime import datetime
from json import dumps

from .version_ import Version, WEB_VERSION

if TYPE_CHECKING:
    from requests.structures import CaseInsensitiveDict
    from werkzeug.datastructures import MultiDict


# Servers may not have a MAX_SOFT_SIZE less than this
MAX_SOFT_SIZE_MIN: int = 8 * (1000**2)


@dataclass
class _ToDict:
    def to_dict(self) -> dict[str, str]:
        return {i.replace("_", "-"): str(k) for i, k in asdict(self).items() if k is not None}


def _get_bool(d: dict[str, str], name: str, default: bool) -> bool:
    return d.get(name, str(default)) == "True"


def _get_int_or_none(d: dict[str, str], name: str) -> int | None:
    got: str | None = d.get(name, None)
    if isinstance(got, str):
        try:
            return int(got)  # If we can't convert to an int, we'll just return None
        except ValueError:
            pass
    return None


#
# Request Params
#


@dataclass(kw_only=True)
class UploadRequestParams(_ToDict):
    version: Version
    encrypted: bool
    final: bool = False
    override: bool = False
    stream_id: str | None = None  # Not required for initial upload POST
    ttl: int | None = None  # Use None if not provided (only read during initial upload POST)

    @classmethod
    def from_dict(cls, d: MultiDict[str, str]) -> UploadRequestParams:
        return cls(
            version=Version(d.get("version", WEB_VERSION.str)),
            encrypted=_get_bool(d, "encrypted", False),
            final=_get_bool(d, "final", False),
            override=_get_bool(d, "override", False),
            stream_id=d.get("stream-id", None),
            ttl=_get_int_or_none(d, "ttl"),
        )


@dataclass(kw_only=True)
class DownloadRequestParams(_ToDict):
    version: Version
    delete: bool
    override: bool = False
    stream_id: str | None = None  # Not required for initial upload GET

    @classmethod
    def from_dict(cls, d: MultiDict[str, str]) -> DownloadRequestParams:
        return cls(
            version=Version(d.get("version", WEB_VERSION.str)),
            delete=_get_bool(d, "delete", False),
            override=_get_bool(d, "override", False),
            stream_id=d.get("stream-id", None),
        )


#
# Response Headers
#


class BadHeaders(RuntimeError):
    """
    Raised a response's headers are bad
    """


_Self = TypeVar("_Self", bound="_ResponseHeaders")  # typing.Self in python3.11


class _ResponseHeaders(_ToDict):
    @classmethod
    def from_dict(cls: type[_Self], d: CaseInsensitiveDict[str]) -> _Self:
        try:
            return cls._from_dict(dict(d.lower_items()))
        except KeyError as e:
            raise BadHeaders("Missing headers") from e

    @classmethod
    def _from_dict(cls: type[_Self], d: dict[str, str]) -> _Self:
        raise NotImplementedError()


@dataclass(kw_only=True)
class UploadResponseHeaders(_ResponseHeaders):
    stream_id: str
    max_size: int

    @classmethod
    def _from_dict(cls, d: dict[str, str]) -> UploadResponseHeaders:
        return cls(stream_id=d["stream-id"], max_size=int(d["max-size"]))


@dataclass(kw_only=True)
class DownloadResponseHeaders(_ResponseHeaders):
    stream_id: str
    final: bool
    encrypted: bool

    @classmethod
    def _from_dict(cls, d: dict[str, str]) -> DownloadResponseHeaders:
        return cls(
            stream_id=d["stream-id"],
            final=d["final"] == "True",
            encrypted=d["encrypted"] == "True",
        )


#
# Other
#


# pylint: disable=too-many-instance-attributes
@dataclass(kw_only=True, frozen=True)
class QueryResponse:
    new: bool
    upload_complete: bool
    packets: int
    size: int
    encrypted: bool
    version: Version
    expiration: datetime
    locked: bool


@dataclass(kw_only=True, frozen=True)
class AdminMessage:
    body: str
    path: str
    uid: str

    def bytes(self) -> bytes:
        return dumps(asdict(self)).encode()



from zstdlib import Enum


BLOCKED_EC: int = 401


# pylint: disable=bad-mcs-method-argument,bad-mcs-classmethod-argument
class UploadEC(Enum):
    """
    HTTP error codes the rpipe client may be sent when uploading data
    Others may be sent, but these are the ones the client should be prepared to handle
    """

    wrong_version: int = 412  #    PUT: different version than initial POST
    illegal_version: int = 426  #  Illegal version
    stream_id: int = 422  #        POST: Has stream ID, should not; PUT: missing stream ID
    too_big: int = 413  #          Too much data sent to server
    conflict: int = 409  #         Stream ID indicates a different stream than exists
    wait: int = 425  #             Try again in a bit, waiting on the other end of the pipe
    forbidden: int = 406  #        Writing to finalized stream
    locked: int = 423  #           Channel is locked and cannot be edited


class DownloadEC(Enum):
    """
    HTTP error codes the rpipe client may be sent when downloading data
    Others may be sent, but these are the ones the client should be prepared to handle
    """

    wrong_version: int = 412  #    GET: bad version
    illegal_version: int = 426  #  Illegal version
    no_data: int = 410  #          No data on this channel; takes priority over stream_id error
    conflict: int = 409  #         Stream ID indicates a different stream than exists
    wait: int = 425  #             Try again in a bit, waiting on the other end of the pipe
    forbidden: int = 406  #        StreamID passed for new stream or while peeking
    cannot_peek: int = 452  #      Cannot peek, too much data
    in_use: int = 453  #           Someone else is reading from the pipe
    locked: int = 423  #           Channel is locked and cannot be edited


class DeleteEC(Enum):
    """
    HTTP error codes the rpipe client may be sent when deleting a channel
    Others may be sent, but these are the ones the client should be prepared to handle
    """

    locked: int = 423  #           Channel is locked and cannot be edited


class QueryEC(Enum):
    """
    HTTP error codes the rpipe client may be sent when in query mode
    Others may be sent, but these are the ones the client should be prepared to handle
    """

    illegal_version: int = 426  #  Illegal version
    no_data: int = 410  #          No data on this channel


class AdminEC(Enum):
    """
    HTTP error codes the rpipe client may be sent when in admin mode
    Others may be sent, but these are the ones the client should be prepared to handle
    """

    invalid: int = 400
    unauthorized: int = 403
    illegal_version: int = 426



from multiprocessing import cpu_count
from pathlib import Path
import argparse

import argcomplete

from .. import __version__  # Extract version without importing shared


PASSWORD_ENV: str = "RPIPE_PASSWORD"
_DEFAULT_CF = Path.home() / ".config" / "rpipe.json"


def _cli(parser: argparse.ArgumentParser, parsed: argparse.Namespace) -> None:
    """
    We import main from CLI after parsing arguments because
    it is slow, and not necessary for --help or --version
    """
    # pylint: disable=import-outside-toplevel,cyclic-import
    from .main import main

    main(parser, parsed)


def _si(size: str) -> int:
    if size.isdecimal():
        return int(size)
    try:
        if (frv := float(size[:-1]) * (1000 ** (1 + "KMGT".index(size[-1].upper())))) == (rv := int(frv)):
            return rv
        raise ValueError("Non-integer number of bytes requested")
    except ValueError as e:
        raise ValueError(f"Invalid size: {size}") from e


# pylint: disable=too-many-locals,too-many-statements
def cli() -> None:
    """
    Parses arguments then invokes rpipe
    """
    threads: int = max(1, (cpu := cpu_count()) - 1)
    parser = argparse.ArgumentParser(add_help=False)
    parser.set_defaults(method=None)
    recv_g = parser.add_argument_group("Recv Mode")
    recv_g.add_argument(
        "-b", "--block", action="store_true", help="Wait until a channel is available to read"
    )
    recv_g.add_argument(
        "-p",
        "--peek",
        action="store_true",
        help="Read pipe without emptying it; will not construct a persistent pipe like a normal read",
    )
    recv_g.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Attempt to read data even if this is a upload/download client version mismatch",
    )
    recv_g.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Overwrite existing output path if it is not a non-empty directory (requires --file or --dir)",
    )
    send_g = parser.add_argument_group("Send Mode")
    send_g.add_argument(
        "-t",
        "--ttl",
        type=int,
        default=None,
        help="Pipe TTL in seconds; use server default if not passed",
    )
    send_g.add_argument(  # Do not use default= for better error checking w.r.t. plaintext mode
        "-Z",
        "--zstd",
        metavar="[1-22]",
        choices=range(1, 23),
        type=int,
        help="Compression level to use; invalid in plaintext mode",
    )
    send_g.add_argument(
        "-j",
        "--threads",
        metavar=f"[1-{cpu}]" if cpu > 1 else "1",
        choices=range(1, cpu + 1),
        default=threads,
        type=int,
        help=f"The number of threads to use for compression. Default: {threads}",
    )
    recv_send_g = parser.add_argument_group("Recv Mode / Send Mode")
    io_g = recv_send_g.add_mutually_exclusive_group()
    io_g.add_argument(
        "-F",
        "--file",
        default=None,
        const=True,
        nargs="?",
        type=Path,
        help="A file use for input/output instead of stdin/stdout. For sending, implies --progress <file size> unless otherwise specified. Requires -r or -s",
    )
    io_g.add_argument(
        "-D",
        "--dir",
        default=None,
        const=True,
        nargs="?",
        type=Path,
        help="A dir to tar/output as input/output instead of stdin/stdout. For sending, implies --progress <tarball size> unless otherwise specified. Requires -r or -s",
    )
    prog_g = recv_send_g.add_mutually_exclusive_group()
    prog_g.add_argument("-N", "--no-progress", action="store_true", help="Do not show a progress bar")
    prog_g.add_argument(
        "-P",
        "--progress",
        metavar="SIZE",
        type=_si,
        default=None,
        const=True,
        nargs="?",
        help="Show a progress bar, if a value is passed, assume that's the number of bytes to be passed. Only valid while sending or receiving data. Values can be suffixed with K, M, G, or T, to multiply by powers of 1000",
    )
    recv_send_g.add_argument(
        "-Y", "--total", action="store_true", help="Print the total number of bytes sent/received"
    )
    recv_send_g.add_argument(
        "-K", "--checksum", action="store_true", help="Checksum the data being sent/received"
    )
    # Config options
    config = parser.add_argument_group("Configuration")
    config.add_argument("-u", "--url", help="The pipe url to use")
    config.add_argument("-c", "--channel", help="The channel to use")
    config.add_argument("-T", "--timeout", type=float, help="The timeout for the HTTP requests")
    config.add_argument(
        "-k",
        "--key-file",
        type=Path,
        help="SSH ed25519 private key file used to signed admin requests",
    )
    config.add_argument("-C", "--config-file", type=Path, default=_DEFAULT_CF, help="The custom config file")
    config.add_argument(
        "-e",
        "--encrypt",
        action=argparse.BooleanOptionalAction,
        help=f"Encrypt the data; uses {PASSWORD_ENV} as the password if set, otherwise uses saved password",
    )
    config.add_argument("-S", "--ssl", action=argparse.BooleanOptionalAction, help="Require host use https")
    log_g = parser.add_argument_group("Logging")
    # pylint: disable=duplicate-code
    log_g.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase Log verbosity, pass more than once to increase verbosity",
    )
    log_g.add_argument("-n", "--no-color-log", action="store_true", help="Disable color in the log output")
    # Modes
    mode_g = parser.add_argument_group(
        "Mode Selection",
        description="Force usage of a specific mode (automatically chosen by default)",
    ).add_mutually_exclusive_group()
    mode_g.add_argument("-r", "--recv", action="store_true", help="Read data from server")
    mode_g.add_argument("-s", "--send", action="store_true", help="Send data to server")
    mode_g.add_argument("-d", "--delete", action="store_true", help="Delete all entries in the channel")
    mode_g.add_argument(
        "-q",
        "--query",
        action="store_true",
        help="Get information on the given channel",
    )
    mode_g.add_argument("-h", "--help", action="help", help="Show this help message and exit")
    mode_g.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{parser.prog} {__version__}",
        help="Show program's version number and exit",
    )
    mode_g.add_argument(
        "-X", "--print-config", action="store_true", help="Print out the config (including CLI args)"
    )
    mode_g.add_argument(
        "-U",
        "--update-config",
        action="store_true",
        help="Update the existing rpipe config then exit; allows incomplete configs to be saved",
    )
    mode_g.add_argument(
        "-O", "--outdated", action="store_true", help="Check if this client is too old for the server"
    )
    mode_g.add_argument("-Q", "--server-version", action="store_true", help="Print the server version")
    mode_g.add_argument(
        "-B", "--blocked", action="store_true", help="Determine if the client is blocked from the server"
    )
    # Top priority mode
    mode_g.add_argument("-A", "--admin", action="store_true", help="Allow use of admin commands")
    # Admin commands
    admin = parser.add_subparsers(
        title="Admin Commands",
        description=(
            "Server admin commands; must be used with --admin and should have a key file set before use."
            " All arguments except logging and config arguments are ignored with admin commands."
            " Server must be configured to accept messages signed by your selected private key file"
        ),
        dest="method",
    )
    admin.add_parser("debug", help="Check if the server is running in debug mode")
    admin.add_parser("channels", help="List all channels with stats")
    admin.add_parser("stats", help="Print various server stats")
    log_p = admin.add_parser("log", help="Download server logs various server stats")
    log_p.add_argument(
        "-o", "--output-file", type=Path, default=None, help="Log output file, instead of stdout"
    )
    log_lvl_p = admin.add_parser("log-level", help="Get/set the server log level")
    log_lvl_p.add_argument("level", default=None, nargs="?", help="The log level for the server to use")
    admin.add_parser("lock", help="Lock the channel")
    admin.add_parser("unlock", help="Unlock the channel")
    for name in ("ip", "route"):
        p2 = admin.add_parser(name, help=f"Block / unblock {name}s, or get a list of blocked {name}s")
        m_g = p2.add_argument_group(
            f"Block / Unblock a given {name}",
            f"If none of these are passed, the command will return the list of banned {name}s",
        )
        m_g.add_argument("--block", nargs="+", help=f"Block a given {name}")
        m_g.add_argument("--unblock", nargs="+", help=f"Unblock a given {name}")
    argcomplete.autocomplete(parser)  # Tab completion
    _cli(parser, parser.parse_args())



from typing import TYPE_CHECKING, Protocol, cast
from logging import getLogger
from base64 import b85decode
from threading import Lock
from pathlib import Path
from json import loads
from time import sleep

from cryptography.hazmat.primitives.serialization import load_ssh_public_key
from cryptography.exceptions import InvalidSignature, UnsupportedAlgorithm
from flask import Response, request

from ...shared import TRACE, AdminMessage, AdminStats, AdminEC, Version, remote_addr
from .uid import UID


if TYPE_CHECKING:
    from ..server import State


MIN_VERSION = Version("9.6.1")


class _Verifier(Protocol):
    def __call__(self, signature: bytes, *, data: bytes) -> None: ...


class Verify:
    """
    A class to manage signature verification of Admin requests
    """

    _BRUTE_FORCE_DELAY = 0.02

    __slots__ = ("uid", "_verifiers", "_log", "_brute_force_lock")

    def __init__(self, key_files: list[Path]):
        self._log = getLogger("Verify")
        self._log.info("Loading signing keys")
        verifiers = {self._load_verifier(k): k for k in key_files}
        _ = verifiers.pop(None, None)
        self._verifiers = cast(dict[_Verifier, Path], verifiers)
        self._brute_force_lock = Lock()
        self.uid = UID()

    def __call__(self, name: str, state: State) -> Response | str:
        try:
            return self._verify(name, state)
        except Exception:  # pylint: disable=broad-except
            self._log.error("Failed due to:", exc_info=True)
            return Response(status=500)

    # Private methods

    def _load_verifier(self, key_file: Path) -> _Verifier | None:
        try:
            if not key_file.exists():
                self._log.error("Key file %s does not exist", key_file)
                return None
            return cast(_Verifier | None, getattr(load_ssh_public_key(key_file.read_bytes()), "verify", None))
        except UnsupportedAlgorithm:
            self._log.error("Skipping %s - Signature verification algorithm not supported", key_file)
            return None

    def _verify_signature(self, signature: bytes, msg: bytes) -> Path | None:
        self._log.debug("Verifying signature of message: %s", msg)
        for fn in self._verifiers:
            try:
                fn(signature, data=msg)
                return self._verifiers[fn]
            except InvalidSignature:
                pass
        return None

    def _verify(self, name: str, state: State) -> Response | str:
        stat = AdminStats(host=remote_addr(), command=name)
        with state as s:
            s.stats.admin.append(stat)
        # Check version
        self._log.debug("Checking version")
        version, post = request.get_data().split(b"\n", 1)
        stat.version = version.decode()
        pth = request.full_path.strip("?")
        if (ver := Version(version)) < MIN_VERSION:
            self._log.warning("Rejecting request; path: %s; client too old: %s < %s", pth, ver, MIN_VERSION)
            _msg = f"Minimum supported client version: {MIN_VERSION}"
            return Response(_msg, status=AdminEC.illegal_version)
        # Extract parameters
        self._log.debug("Extracting request signature and message")
        signature, msg_bytes = post.split(b"\n", 1)
        msg = AdminMessage(**loads(msg_bytes.decode()))
        stat.uid = msg.uid
        # Verify UID
        with self._brute_force_lock:  # Slow down brute force attacks
            self._log.log(TRACE, "Sleeping %s seconds to prevent brute forcing", self._BRUTE_FORCE_DELAY)
            sleep(self._BRUTE_FORCE_DELAY)
        if not self.uid.verify(msg.uid):
            self._log.error("Rejecting request; invalid UID: %s and path: %s", pth, msg.uid)
            return Response(status=AdminEC.unauthorized)
        stat.uid_valid = True
        # Verify signature
        if (key_file := self._verify_signature(b85decode(signature), msg_bytes)) is None:
            self._log.error("Signature verification failed; path: %s", pth)
            return Response(status=AdminEC.unauthorized)
        stat.signer = key_file
        # Success
        self._log.debug("Signature verified")
        self._log.info("Executing %s", pth)
        return msg.body



from typing import TYPE_CHECKING
from logging import getLogger
from threading import Thread
from time import sleep

from ...shared import TRACE

if TYPE_CHECKING:
    from .state import State


class PruneThread(Thread):
    """
    A thread class that periodically prunes expired pipes from the server
    """

    def __init__(self, state: State):
        super().__init__(target=self._periodic_prune, daemon=True)
        self._state: State = state

    def _periodic_prune(self) -> None:
        log = getLogger("Prune Thread")
        log.info("Starting prune loop")
        while True:
            log.log(TRACE, "Acquiring state lock")
            with self._state as rw_state:
                if rw_state.shutdown:
                    log.debug("Quitting, state is shutdown")
                    return
                log.log(TRACE, "Checking for expired streams")
                expired = [i for i, k in rw_state.streams.items() if k.expired()]
                log.log(TRACE, "Pruning %d expired streams", len(expired))
                for i in expired:
                    log.info("Pruning expired channel %s", i)
                    del rw_state.streams[i]
            log.log(TRACE, "Sleeping for 5 seconds")
            sleep(5)  # Wait a few seconds before checking again
